// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file PeriodicBC.h */

#ifndef AMDIS_PERIODICBC_H
#define AMDIS_PERIODICBC_H

#include <map>
#include <vector>
#include "AMDiS_fwd.h"
#include "BoundaryCondition.h"
#include "FixVec.h"

namespace AMDiS {

  template<typename T>
  class DimVecLess 
  {
  public:
    bool operator()(const DimVec<T> &v1, const DimVec<T> &v2) 
    {
      int  size = v1.getSize();
      for (int i = 0; i < size; i++) {
	if (v1[i] < v2[i]) 
	  return true;
	if (v1[i] > v2[i]) 
	  return false;
      }
      return false;
    }
  };

  class PeriodicDOFMapping
  {
  public:
    PeriodicDOFMapping(const BasisFunction *basFcts);

    ~PeriodicDOFMapping();

  public:
    static PeriodicDOFMapping* providePeriodicDOFMapping(const BasisFunction *basFcts);

    const DegreeOfFreedom *getDOFPermutation(const DimVec<int> &vertexPermutation);

  protected:
    const BasisFunction *basFcts_;

    std::map<DimVec<int>, DegreeOfFreedom*, DimVecLess<int> > dofPermutation_;

    std::map<DimVec<double>, int, DimVecLess<double> > indexOfCoords_;

    static std::vector<PeriodicDOFMapping*> mappings_;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * Sub class of BoundaryCondition. Implements Periodic boundary conditions.
   */
  class PeriodicBC : public BoundaryCondition
  {
  public:
    /// Constructor.
    PeriodicBC(BoundaryType type, FiniteElemSpace *rowFESpace_);

    ~PeriodicBC();

    void initMatrix(DOFMatrix* matrix);

    void fillBoundaryCondition(DOFMatrix *matrix,
			       ElInfo *elInfo,
			       const DegreeOfFreedom *dofIndices,
			       const BoundaryType *localBound,
			       int nBasFcts);

    void exitMatrix(DOFMatrix* matrix);

    void exitVector(DOFVectorBase<double>* vector);

  protected:
    VertexVector *associated;

    PeriodicDOFMapping *periodicDOFMapping_;

    DegreeOfFreedom *neighIndices_;

    DOFMatrix *masterMatrix_;
  };

}

#endif
